import QtQuick 1.0
import QtMultimediaKit 1.1
import Unity2d 1.0

import "../utils.js" as Utils

FocusScope {
    property bool playing: state == "playing"
    property bool paused: state == "paused"
    property real aspectRatio: width / height
    property alias source: video.source
    property alias status: video.status

    id: player
    state: "stopped"

    function stop() {
        state = "stopped"
        source = ""
    }

    function play() {
        state = "playing"
    }

    function pause() {
        state = "paused"
    }

    function togglePause() {
        if (playing) {
            if (paused) { play() }
            else { pause() }
        }
    }

    function seekForward() {
        return seek(60000)
    }

    function seekBackward() {
        return seek(-15000)
    }

    function seek(value) {
        if (!video.paused) triggerControls.restart()
        if (video.seekable) {
            if (state != "playing" && state != "paused") {
                state = "playing"
                state = "paused"
            }
            video.position += value
        } else { return false }
        return true
    }

    function fastForward() {
        if (video.seekable) {
            if (state == "forwarding") {
                state = "fastForwarding"
            } else if (state == "fastForwarding") {
                scrubbingTimer.step = Math.min(60000, scrubbingTimer.step * 3)
            } else {
                state = "forwarding"
            }
            return true
        } else { return false }
    }

    function rewind() {
        if (video.seekable) {
            if (state == "rewinding") {
                scrubbingTimer.step = Math.max(-60000, scrubbingTimer.step * 3)
            } else {
                state = "rewinding"
            }
            return true
        } else {
            return false
        }
    }

    Rectangle {
        id: playerBackground
        anchors.fill: parent
        color: "black"
    }

    Video {
        property real aspectRatio: metaData.resolution !== undefined ? metaData.resolution.width / metaData.resolution.height : -1
        property int realWidth: aspectRatio != -1 ? Math.min(width, (aspectRatio >= parent.aspectRatio ? width : aspectRatio * height)) : -1
        property int realHeight: aspectRatio != -1 ? Math.min(height, (aspectRatio < parent.aspectRatio ? height : width / aspectRatio)) : -1
        property int widthMargin: aspectRatio != -1 ? (aspectRatio >= parent.aspectRatio ? 0 : (width - realWidth) / 2) : -1
        property int heightMargin: aspectRatio != -1 ? (aspectRatio < parent.aspectRatio ? 0 : (height - realHeight) / 2) : -1
        property real lastVolume
        id: video
        anchors.fill: parent
        smooth: true

        onMutedChanged:
            // FIXME: work around muted not actually muting
            if (muted) {
                lastVolume = volume
                volume = 0
            } else {
                volume = lastVolume
            }
    }

    Timer {
        id: scrubbingTimer
        interval: 500
        repeat: true

        property int step

        onTriggered:
            if (video.position + step < 0) {
                video.position = 0
                player.state = "playing"
            } else {
                video.position += step
            }
    }

    Text {
        id: title
        anchors.left: parent.left
        anchors.leftMargin: video.widthMargin + video.realWidth * 0.05
        anchors.right: parent.right
        anchors.rightMargin: video.widthMargin + video.realWidth * 0.05
        anchors.top: parent.top
        anchors.topMargin: video.heightMargin + video.realWidth * 0.05

        font.family: "Ubuntu"
        font.pixelSize: video.realWidth * 0.04
        color: "white"
        style: Text.Outline
        styleColor: "grey"

        elide: video.metaData.title !== undefined ? Text.ElideMiddle : Text.ElideLeft
        text: video.metaData.title !== undefined ? video.metaData.title : video.source
    }

    Item {
        id: controls
        anchors.left: parent.left
        anchors.right: parent.right
        anchors.bottom: parent.bottom
        height: 160

        Rectangle {
            id: timelineBackground
            anchors.fill: parent
            color: Utils.darkAubergine
            opacity: 0.56
        }

        BorderImage {
            id: timeline
            source: "artwork/timeline.png"

            anchors.verticalCenter: parent.verticalCenter
            anchors.left: parent.left; anchors.leftMargin: 48
            anchors.right: parent.right; anchors.rightMargin: 268
            height: 36

            border.left: 17; border.top: 18
            border.right: 17; border.bottom: 18
        }

        Item {
            id: timelineContents
            anchors.verticalCenter: parent.verticalCenter
            anchors.left: timeline.left; anchors.leftMargin: 10
            anchors.right: timeline.right; anchors.rightMargin: 10
            height: 18

            Item {
                id: bufferFiller
                anchors.top: parent.top; anchors.bottom: parent.bottom
                anchors.left: parent.left
                // FIXME make the buffer bar reflect buffer fill
                width: 0
                clip: true

                Behavior on width { NumberAnimation { duration: 125 } }

                BorderImage {
                        id: bufferFillerImage
                        source: "artwork/buffer_filler.png"

                        anchors.left: parent.left
                        anchors.top: parent.top
                        anchors.bottom: parent.bottom
                        width: timelineContents.width

                        border.left: 8; border.top: 9
                        border.right: 8; border.bottom: 9
                }
            }

            Item {
                id: positionFiller
                width: timelineContents.width * (video.position / video.duration)
                anchors.top: parent.top; anchors.bottom: parent.bottom
                anchors.left: parent.left
                clip: true

                Behavior on width { NumberAnimation { duration: 125 } }

                BorderImage {
                        id: positionFillerImage
                        source: "artwork/position_filler.png"

                        anchors.left: parent.left
                        anchors.top: parent.top
                        anchors.bottom: parent.bottom
                        width: timelineContents.width

                        border.left: 8; border.top: 9
                        border.right: 8; border.bottom: 9
                }
            }

/*          // FIXME need to investigate the glow getting offset when overall glow is enabled
            GlowEffect {
                id: buttonGlow
                sourceItem: button
            }
*/
            AbstractButton {
                id: button
                anchors.horizontalCenter: positionFiller.right
                anchors.verticalCenter: parent.verticalCenter
                width: 100; height: width
                focus: true

                property string iconPath: "artwork/icon_%1.png"
                property string icon
                property bool switchIcons: false

                Rectangle {
                    id: buttonBackground
                    anchors.fill: parent
                    anchors.margins: parent.width * 0.15
                    color: "#482d2a"
                    radius: 18

                    border.color: "#ffffff"
                    border.width: width * 0.04
                }

                GlowEffect {
                    id: iconOneGlow
                    sourceItem: iconOne
                    overflow: 0
                    blur: 1.5
                    opacity: iconOne.opacity
                }

                Image {
                    id: iconOne
                    anchors.centerIn: parent
                    width: 64; height: width
                    smooth: true
                    opacity: button.switchIcons ? 0 : 1

                    Behavior on opacity { NumberAnimation { property: "opacity"; duration: 125 } }
                }

                ColorizeEffect {
                    id: iconOneColorize
                    sourceItem: iconOne
                    color: Utils.orange
                }

                GlowEffect {
                    id: iconTwoGlow
                    sourceItem: iconTwo
                    overflow: 0
                    blur: 1.5
                    opacity: iconTwo.opacity
                }

                Image {
                    id: iconTwo
                    anchors.centerIn: parent
                    width: 64; height: width
                    smooth: true
                    opacity: button.switchIcons ? 1 : 0

                    Behavior on opacity { NumberAnimation { property: "opacity"; duration: 125 } }
                }

                ColorizeEffect {
                    id: iconTwoColorize
                    sourceItem: iconTwo
                    color: Utils.orange
                }

                onIconChanged: {
                    if (!switchIcons) { iconTwo.source = iconPath.arg(icon) }
                    else { iconOne.source = iconPath.arg(icon) }
                    switchIcons = !switchIcons
                }

                onClicked: player.togglePause()

                states: [
                    State {
                        name: "selected"
                        PropertyChanges { target: iconOneColorize; }
                        PropertyChanges { target: iconTwoColorize; }
                    },

                    State {
                        name: "default"
                        PropertyChanges { target: iconOneColorize; opacity: 0; enabled: false }
                        PropertyChanges { target: iconTwoColorize; opacity: 0; enabled: false }
                    }
                ]
            }
        }

        Item {
            anchors.left: timeline.right
            anchors.right: parent.right
            anchors.top: parent.top
            anchors.bottom: parent.bottom

            Text {
                id: positionText
                anchors.centerIn: parent

                font.family: "Ubuntu"
                font.pixelSize: 35
                color: "white"
                style: Text.Outline
                styleColor: "grey"

                text: Utils.format_time(video.position, Utils.get_human_time_format(video.duration))
            }
        }
    }

    states: [
        State {
            name: "stopped"
            PropertyChanges { target: video; playing: false; paused: false }
            PropertyChanges { target: title; opacity: 0 }
            PropertyChanges { target: scrubbingTimer; running: false }
            PropertyChanges { target: button; icon: "stop" }
        },

        State {
            name: "playing"
            PropertyChanges { target: video; playing: true; paused: false; playbackRate: 1.0; muted: false}
            PropertyChanges { target: title; opacity: video.hasVideo ? 0 : 1 }
            PropertyChanges { target: controls; opacity: video.hasVideo ? 0 : 1 }
            PropertyChanges { target: scrubbingTimer; running: false }
            PropertyChanges { target: button; icon: "play" }
        },

        State {
            name: "paused"
            PropertyChanges { target: video; playing: true; paused: true }
            PropertyChanges { target: title; opacity: 1 }
            PropertyChanges { target: controls; opacity: 1 }
            PropertyChanges { target: scrubbingTimer; running: false }
            PropertyChanges { target: button; icon: "pause" }
        },

        State {
            name: "forwarding"
            PropertyChanges { target: video; playing: true; paused: false; playbackRate: 2.0; muted: true }
            PropertyChanges { target: title; opacity: video.hasVideo ? 0 : 1 }
            PropertyChanges { target: controls; opacity: 1 }
            PropertyChanges { target: scrubbingTimer; running: false }
            PropertyChanges { target: button; icon: "forward" }
         },

        State {
            name: "fastForwarding"
            PropertyChanges { target: video; playing: true; paused: false; muted: true }
            PropertyChanges { target: title; opacity: video.hasVideo ? 0 : 1 }
            PropertyChanges { target: controls; opacity: 1 }
            PropertyChanges { target: scrubbingTimer; running: true }
            PropertyChanges { target: button; icon: "forward" }
        },

        State {
            name: "rewinding"
            PropertyChanges { target: video; playing: true; paused: false; muted: true }
            PropertyChanges { target: title; opacity: video.hasVideo ? 0 : 1 }
            PropertyChanges { target: controls; opacity: 1 }
            PropertyChanges { target: scrubbingTimer; running: true }
            PropertyChanges { target: button; icon: "rewind" }
        }
    ]

    transitions: [
        Transition {
            to: "stopped"
            NumberAnimation { target: title; property: "opacity"; duration: 125 }
            NumberAnimation { target: controls; property: "opacity"; duration: 125 }
        },

        Transition {
            to: "paused"
            PropertyAction { target: triggerControls; property: "running"; value: false }
            NumberAnimation { target: title; property: "opacity"; duration: 125 }
            NumberAnimation { target: controls; property: "opacity"; duration: 125; to: 1 }
        },

        Transition {
            from: "stopped"
            to: "playing"

            animations: [
                SequentialAnimation {
                    NumberAnimation { target: title; duration: 125; property: "opacity"; to: 1 }
                    NumberAnimation { duration: 2000 }
                    NumberAnimation { target: title; property: "opacity"; duration: 500 }
                },

                SequentialAnimation {
                    NumberAnimation { duration: 1500 }
                    NumberAnimation { target: controls; property: "opacity"; duration: 125 }
                }
            ]
        },

        Transition {
            to: "playing"

            animations: [
                SequentialAnimation {
                    NumberAnimation { duration: 1000 }
                    NumberAnimation { target: controls; property: "opacity"; duration: 125 }
                }
            ]
        },

        Transition {
            to: "forwarding"
            PropertyAction { target: triggerControls; property: "running"; value: false }
            NumberAnimation { target: controls; property: "opacity" }
        },

        Transition {
            to: "fastForwarding"
            PropertyAction { target: triggerControls; property: "running"; value: false }
            NumberAnimation { target: controls; property: "opacity" }
            PropertyAction { target: scrubbingTimer; property: "step"; value: 1000 }
        },

        Transition {
            to: "rewinding"
            PropertyAction { target: triggerControls; property: "running"; value: false }
            NumberAnimation { target: controls; property: "opacity" }
            PropertyAction { target: scrubbingTimer; property: "step"; value: -500 }
        }
    ]

    SequentialAnimation {
        id: triggerControls
        NumberAnimation { target: controls; property: "opacity"; duration: 125; to: 1 }
        NumberAnimation { duration: 3000 }
        NumberAnimation { target: controls; property: "opacity"; duration: 125; to: 0 }
    }
}
